import os
import tempfile
from typing import Hashable, Literal, TypeAlias

import numpy as np
import xarray as xr

from clouddrift.adapters.utils import download_with_progress
from clouddrift.raggedarray import RaggedArray

_DEFAULT_FILE_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", "ibtracs")

_SOURCE_BASE_URI = "https://www.ncei.noaa.gov/data/international-best-track-archive-for-climate-stewardship-ibtracs"

_SOURCE_URL_FMT = "{base_uri}/{version}/access/netcdf/IBTrACS.{kind}.{version}.nc"

_Version: TypeAlias = Literal["v04r00"] | Literal["v04r01"]

_Kind: TypeAlias = (
    Literal["ACTIVE"]
    | Literal["ALL"]
    | Literal["EP"]
    | Literal["NA"]
    | Literal["NI"]
    | Literal["SA"]
    | Literal["SI"]
    | Literal["SP"]
    | Literal["WP"]
    | Literal["LAST_3_YEARS"]
    | Literal["SINCE_1980"]
)


def to_raggedarray(
    version: _Version, kind: _Kind, tmp_path: str = _DEFAULT_FILE_PATH
) -> xr.Dataset:
    """Returns International Best Track Archive for Climate Stewardship (IBTrACS) as a ragged array xarray dataset.

    The upstream data is available at https://www.ncei.noaa.gov/products/international-best-track-archive

    Parameters
    ----------
    version : str, optional
        Specify the dataset version to retrieve. Default to the latest version. Default is "v04r01".
    kind: str, optional
        Specify the dataset kind to retrieve. Specifying the kind can speed up execution time of specific querries
        and operations. Default is "LAST_3_YEARS".
    tmp_path: str, default adapter temp path (default)
        Temporary path where intermediary files are stored. Default is ${osSpecificTempFileLocation}/clouddrift/ibtracs/.

    Returns
    -------
    xarray.Dataset
        IBTRACS dataset as a ragged array.
    """
    ds = _get_original_dataset(version, kind, tmp_path)
    ds = ds.rename_dims({"date_time": "obs"})

    vars = list[Hashable]()
    vars.extend(ds.variables.keys())
    for coord in ds.coords.keys():
        vars.remove(coord)
    dtypes = {v: ds[v].dtype for v in vars}
    dtypes.update({"numobs": np.dtype("int64")})
    ds = ds.astype(dtypes)

    data_vars = list()
    md_vars = list()

    for var_name in ds.variables:
        # time variable shouldn't be considered a data or metadata variable
        if var_name in ["time"]:
            continue

        var: xr.DataArray = ds[var_name]

        if "obs" in var.dims and len(var.dims) >= 2:
            data_vars.append(var_name)
        elif len(var.dims) == 1 and var.dims[0] == "storm":
            md_vars.append(var_name)

    ra = RaggedArray.from_files(
        indices=list(range(0, len(ds["sid"]))),
        name_coords=["id", "time"],
        name_meta=md_vars,
        name_data=data_vars,
        name_dims={"storm": "rows", "obs": "obs", "quadrant": "quadrant"},
        rowsize_func=_rowsize,
        preprocess_func=_preprocess,
        attrs_global=ds.attrs,
        attrs_variables={
            var_name: ds[var_name].attrs for var_name in data_vars + md_vars
        },
        dataset=ds,
        data_vars=data_vars,
        md_vars=md_vars,
    )
    return ra.to_xarray()


def _get_original_dataset(
    version: _Version, kind: _Kind, tmp_path: str = _DEFAULT_FILE_PATH
) -> xr.Dataset:
    os.makedirs(tmp_path, exist_ok=True)
    src_url = _get_source_url(version, kind)

    filename = src_url.split("/")[-1]
    dst_path = os.path.join(tmp_path, filename)
    download_with_progress([(src_url, dst_path)])

    return xr.open_dataset(dst_path, engine="netcdf4")


def _rowsize(idx: int, **kwargs):
    ds: xr.Dataset | None = kwargs.get("dataset")
    if ds is None:
        raise ValueError("kwargs dataset missing")
    storm_ds = ds.isel(storm=idx)
    return storm_ds["numobs"].data


def _preprocess(idx: int, **kwargs):
    ds: xr.Dataset | None = kwargs.get("dataset")
    data_vars: list[str] | None = kwargs.get("data_vars")
    md_vars: list[str] | None = kwargs.get("md_vars")

    if ds is not None and data_vars is not None and md_vars is not None:
        storm_ds = ds.isel(storm=idx)
        numobs = storm_ds["numobs"].data
        vars = dict()

        for var in data_vars + list(storm_ds.coords):
            if var != "time":
                vars.update({var: (storm_ds[var].dims, storm_ds[var].data[:numobs])})

        for var in md_vars:
            vars.update({var: (("storm",), [storm_ds[var].data])})

        return xr.Dataset(
            vars,
            {
                "id": (("storm",), np.array([idx])),
                "time": (("obs",), storm_ds["time"].data[:numobs]),
            },
        )
    else:
        raise ValueError("kwargs dataset, data vars and md_vars missing")


def _kind_map(kind: _Kind):
    return {"LAST_3_YEARS": "last3years", "SINCE_1980": "since1980"}.get(kind, kind)


def _get_source_url(version: _Version, kind: _Kind):
    return _SOURCE_URL_FMT.format(
        base_uri=_SOURCE_BASE_URI, version=version, kind=_kind_map(kind)
    )



import os
import shutil

import numpy as np

import tests.utils as testutils
from clouddrift.adapters import gdp1h


class gdp1h_integration_tests(testutils.DisableProgressTestCase):
    def test_load_subset_and_create_aggregate(self):
        test_tasks = [
            (gdp1h.GDP_TMP_PATH, gdp1h.GDP_DATA_URL),
            (gdp1h.GDP_TMP_PATH_EXPERIMENTAL, gdp1h.GDP_DATA_URL_EXPERIMENTAL),
        ]

        for path, url in test_tasks:
            with self.subTest(
                f"test downloading and creating ragged array for: ({url})"
            ):
                ra = gdp1h.to_raggedarray(
                    drifter_ids=[
                        63123,
                        92905,
                        101877,
                        300234060218770,
                        300234062951460,
                    ],
                    tmp_path=path,
                    url=url,
                )

                assert "rowsize" in ra.metadata
                assert "sst" in ra.data
                assert "ve" in ra.data
                assert "vn" in ra.data
                assert ra.coords["id"].dtype == np.int64
                assert len(ra.data["vn"]) == len(ra.coords["time"])
                assert len(ra.data["ve"]) == len(ra.coords["time"])
                assert len(ra.data["sst"]) == len(ra.coords["time"])
                assert len(ra.metadata["rowsize"]) == len(ra.coords["id"])
                self.assertTrue(np.any(~ra.metadata["location_type"]))

                agg_path = os.path.join(path, "aggregate")
                os.makedirs(agg_path, exist_ok=True)
                ra.to_netcdf(os.path.join(agg_path, "gdp1h_5r_sample.nc"))

    @classmethod
    def tearDownClass(cls):
        [
            shutil.rmtree(dir)
            for dir in [gdp1h.GDP_TMP_PATH, gdp1h.GDP_TMP_PATH_EXPERIMENTAL]
        ]



from ._version import version

__version__ = version

import clouddrift.adapters as adapters
import clouddrift.datasets as datasets
import clouddrift.kinematics as kinematics
import clouddrift.pairs as pairs
import clouddrift.plotting as plotting
import clouddrift.ragged as ragged
import clouddrift.signal as signal
import clouddrift.sphere as sphere
import clouddrift.transfer as transfer
import clouddrift.wavelet as wavelet
from clouddrift.raggedarray import RaggedArray

__all__ = [
    "RaggedArray",
    "adapters",
    "datasets",
    "kinematics",
    "pairs",
    "plotting",
    "ragged",
    "signal",
    "sphere",
    "transfer",
    "wavelet",
]



"""
This module provides functions and metadata that can be used to convert the
hourly Global Drifter Program (GDP) data to a ``clouddrift.RaggedArray``
instance.
"""

import logging
import os
import re
import tempfile
import urllib.request
import warnings
from collections.abc import Sequence
from datetime import datetime, timedelta

import numpy as np
import xarray as xr

import clouddrift.adapters.gdp as gdp
from clouddrift.adapters.utils import download_with_progress, standard_retry_protocol
from clouddrift.raggedarray import RaggedArray

GDP_VERSION = "2.01.1"


GDP_DATA_URL = "https://www.aoml.noaa.gov/ftp/pub/phod/buoydata/hourly_product/v2.01"
GDP_DATA_URL_EXPERIMENTAL = (
    "https://www.aoml.noaa.gov/ftp/pub/phod/lumpkin/hourly/experimental"
)


GDP_TMP_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", "gdp")
GDP_TMP_PATH_EXPERIMENTAL = os.path.join(tempfile.gettempdir(), "clouddrift", "gdp_exp")
GDP_DATA = [
    "lon",
    "lat",
    "ve",
    "vn",
    "err_lat",
    "err_lon",
    "err_ve",
    "err_vn",
    "gap",
    "sst",
    "sst1",
    "sst2",
    "err_sst",
    "err_sst1",
    "err_sst2",
    "flg_sst",
    "flg_sst1",
    "flg_sst2",
    "drogue_status",
]

_logger = logging.getLogger(__name__)


def download(
    url: str = GDP_DATA_URL,
    tmp_path: str = GDP_TMP_PATH,
    drifter_ids: list[int] | None = None,
    n_random_id: int | None = None,
):
    """Download individual NetCDF files from the AOML server.

    Parameters
    ----------
    url : str
        URL from which to download the data.
    tmp_path : str
        Path to the directory where the individual NetCDF files are stored.

    drifter_ids : list, optional
        List of drifter to retrieve (Default: all)
    n_random_id : int, optional
        Randomly select n_random_id drifter IDs to download (Default: None)
    Returns
    -------
    out : list
        List of retrieved drifters
    """
    _logger.debug(f"Downloading GDP hourly data from ({url}) to ({tmp_path})")

    # Create a temporary directory if doesn't already exists.
    os.makedirs(tmp_path, exist_ok=True)
    pattern = "drifter_hourly_[0-9]*.nc"
    filename_pattern = "drifter_hourly_{id}.nc"

    # retrieve all drifter ID numbers
    if drifter_ids is None:
        urlpath = standard_retry_protocol(lambda: urllib.request.urlopen(url))()
        string = urlpath.read().decode("utf-8")
        filelist: Sequence[str] = re.compile(pattern).findall(string)  # noqa: F821
    else:
        filelist = [filename_pattern.format(id=did) for did in drifter_ids]
    filelist = list(np.unique(filelist))

    # retrieve only a subset of n_random_id trajectories
    if n_random_id:
        if n_random_id > len(filelist):
            _logger.warn(
                f"Retrieving all listed trajectories because {n_random_id} is larger than the {len(filelist)} listed trajectories."
            )
        else:
            rng = np.random.Generator(np.random.MT19937(42))
            filelist = sorted(rng.choice(filelist, n_random_id, replace=False))

    download_with_progress(
        [(f"{url}/{f}", os.path.join(tmp_path, f)) for f in filelist]
    )
    # Download the metadata so we can order the drifter IDs by end date.
    gdp_metadata = gdp.get_gdp_metadata(tmp_path)

    return gdp.order_by_date(
        gdp_metadata, [int(f.split("_")[-1].removesuffix(".nc")) for f in filelist]
    )


def preprocess(index: int, **kwargs) -> xr.Dataset:
    """Extract and preprocess the Lagrangian data and attributes.

    This function takes an identification number that can be used to create a
    file or url pattern or select data from a Dataframe. It then preprocesses
    the data and returns a clean Xarray Dataset.

    Parameters
    ----------
    index : int
        Drifter's identification number

    Returns
    -------
    ds : xr.Dataset
        Xarray Dataset containing the data and attributes
    """
    fp = os.path.join(kwargs["tmp_path"], kwargs["filename_pattern"].format(id=index))
    ds = xr.open_dataset(
        fp,
        decode_times=False,
        decode_coords=False,
        engine="netcdf4",
    )

    # parse the date with custom function
    ds["deploy_date"].data = gdp.decode_date(np.array([ds.deploy_date.data[0]]))
    ds["end_date"].data = gdp.decode_date(np.array([ds.end_date.data[0]]))
    ds["drogue_lost_date"].data = gdp.decode_date(
        np.array([ds.drogue_lost_date.data[0]])
    )
    ds["time"].data = gdp.decode_date(np.array([ds.time.data[0]]))

    # convert fill values to nan
    for var in [
        "err_lon",
        "err_lat",
        "err_ve",
        "err_vn",
        "sst",
        "sst1",
        "sst2",
        "err_sst",
        "err_sst1",
        "err_sst2",
    ]:
        try:
            ds[var].data = gdp.fill_values(ds[var].data)
        except KeyError:
            warnings.warn(f"Variable {var} not found; skipping.")

    # fix missing values stored as str
    for var in [
        "longitude",
        "latitude",
        "err_lat",
        "err_lon",
        "ve",
        "vn",
        "err_ve",
        "err_vn",
        "sst",
        "sst1",
        "sst2",
    ]:
        try:
            ds[var].encoding["missing value"] = -1e-34
        except KeyError:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")

    # convert type of some variable
    target_dtype = {
        "ID": "int64",
        "WMO": "int32",
        "expno": "int32",
        "typedeath": "int8",
        "flg_sst": "int8",
        "flg_sst1": "int8",
        "flg_sst2": "int8",
    }

    for var in target_dtype.keys():
        if var in ds.keys():
            ds[var].data = ds[var].data.astype(target_dtype[var])
        else:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")

    # new variables
    ds["drogue_status"] = (
        ["traj", "obs"],
        [gdp.drogue_presence(ds.drogue_lost_date.data, ds.time.data[0])],
    )

    # convert attributes to variable
    ds["location_type"] = (
        ("traj"),
        [False if ds.attrs.get("location_type") == "Argos" else True],
    )  # 0 for Argos, 1 for GPS
    ds["DeployingShip"] = (("traj"), gdp.cut_str(ds.attrs.get("DeployingShip", ""), 20))
    ds["DeploymentStatus"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("DeploymentStatus", ""), 20),
    )
    ds["BuoyTypeManufacturer"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("BuoyTypeManufacturer", ""), 20),
    )
    ds["BuoyTypeSensorArray"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("BuoyTypeSensorArray", ""), 20),
    )
    ds["CurrentProgram"] = (
        ("traj"),
        np.array(
            [gdp.str_to_float(ds.attrs.get("CurrentProgram", ""), -1)], dtype=np.int32
        ),
    )
    ds["PurchaserFunding"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("PurchaserFunding", ""), 20),
    )
    ds["SensorUpgrade"] = (("traj"), gdp.cut_str(ds.attrs.get("SensorUpgrade", ""), 20))
    ds["Transmissions"] = (("traj"), gdp.cut_str(ds.attrs.get("Transmissions", ""), 20))
    ds["DeployingCountry"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("DeployingCountry", ""), 20),
    )
    ds["DeploymentComments"] = (
        ("traj"),
        np.array(
            [gdp.str_to_float(ds.attrs.get("ManufactureYear", ""), -1)], dtype=np.int16
        ),
    )
    ds["ManufactureMonth"] = (
        ("traj"),
        np.array(
            [gdp.str_to_float(ds.attrs.get("ManufactureMonth", ""), -1)], dtype=np.int16
        ),
    )
    ds["ManufactureSensorType"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("ManufactureSensorType", ""), 20),
    )
    ds["ManufactureVoltage"] = (
        ("traj"),
        np.array(
            [gdp.str_to_float(ds.attrs.get("ManufactureVoltage", "")[:-2], -1)],
            dtype=np.int16,
        ),
    )  # e.g. 56 V
    ds["FloatDiameter"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("FloatDiameter", "")[:-3])],
    )  # e.g. 35.5 cm
    ds["SubsfcFloatPresence"] = (
        ("traj"),
        np.array(
            [gdp.str_to_float(ds.attrs.get("SubsfcFloatPresence", ""))], dtype="bool"
        ),
    )
    ds["DrogueType"] = (("traj"), gdp.cut_str(ds.attrs.get("DrogueType", ""), 7))
    ds["DrogueLength"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("DrogueLength", "")[:-2])],
    )  # e.g. 4.8 m
    ds["DrogueBallast"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("DrogueBallast", "")[:-3])],
    )  # e.g. 1.4 kg
    ds["DragAreaAboveDrogue"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("DragAreaAboveDrogue", "")[:-4])],
    )  # 10.66 m^2
    ds["DragAreaOfDrogue"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("DragAreaOfDrogue", "")[:-4])],
    )  # e.g. 416.6 m^2
    ds["DragAreaRatio"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("DragAreaRatio", ""))],
    )  # e.g. 39.08
    ds["DrogueCenterDepth"] = (
        ("traj"),
        [gdp.str_to_float(ds.attrs.get("DrogueCenterDepth", "")[:-2])],
    )  # e.g. 20.0 m
    ds["DrogueDetectSensor"] = (
        ("traj"),
        gdp.cut_str(ds.attrs.get("DrogueDetectSensor", ""), 20),
    )

    # vars attributes
    vars_attrs = {
        "ID": {"long_name": "Global Drifter Program Buoy ID", "units": "-"},
        "longitude": {"long_name": "Longitude", "units": "degrees_east"},
        "latitude": {"long_name": "Latitude", "units": "degrees_north"},
        "time": {"long_name": "Time", "units": "seconds since 1970-01-01 00:00:00"},
        "rowsize": {
            "long_name": "Number of observations per trajectory",
            "sample_dimension": "obs",
            "units": "-",
        },
        "location_type": {
            "long_name": "Satellite-based location system",
            "units": "-",
            "comments": "0 (Argos), 1 (GPS)",
        },
        "WMO": {
            "long_name": "World Meteorological Organization buoy identification number",
            "units": "-",
        },
        "expno": {"long_name": "Experiment number", "units": "-"},
        "deploy_date": {
            "long_name": "Deployment date and time",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "deploy_lon": {"long_name": "Deployment longitude", "units": "degrees_east"},
        "deploy_lat": {"long_name": "Deployment latitude", "units": "degrees_north"},
        "start_date": {
            "long_name": "First good date and time derived by DAC quality control",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "start_lon": {
            "long_name": "First good longitude derived by DAC quality control",
            "units": "degrees_east",
        },
        "start_lat": {
            "long_name": "Last good latitude derived by DAC quality control",
            "units": "degrees_north",
        },
        "end_date": {
            "long_name": "Last good date and time derived by DAC quality control",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "end_lon": {
            "long_name": "Last good longitude derived by DAC quality control",
            "units": "degrees_east",
        },
        "end_lat": {
            "long_name": "Last good latitude derived by DAC quality control",
            "units": "degrees_north",
        },
        "drogue_lost_date": {
            "long_name": "Date and time of drogue loss",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "typedeath": {
            "long_name": "Type of death",
            "units": "-",
            "comments": "0 (buoy still alive), 1 (buoy ran aground), 2 (picked up by vessel), 3 (stop transmitting), 4 (sporadic transmissions), 5 (bad batteries), 6 (inactive status)",
        },
        "typebuoy": {
            "long_name": "Buoy type (see https://www.aoml.noaa.gov/phod/dac/dirall.html)",
            "units": "-",
        },
        "DeployingShip": {"long_name": "Name of deployment ship", "units": "-"},
        "DeploymentStatus": {"long_name": "Deployment status", "units": "-"},
        "BuoyTypeManufacturer": {"long_name": "Buoy type manufacturer", "units": "-"},
        "BuoyTypeSensorArray": {"long_name": "Buoy type sensor array", "units": "-"},
        "CurrentProgram": {
            "long_name": "Current Program",
            "units": "-",
            "_FillValue": "-1",
        },
        "PurchaserFunding": {"long_name": "Purchaser funding", "units": "-"},
        "SensorUpgrade": {"long_name": "Sensor upgrade", "units": "-"},
        "Transmissions": {"long_name": "Transmissions", "units": "-"},
        "DeployingCountry": {"long_name": "Deploying country", "units": "-"},
        "DeploymentComments": {"long_name": "Deployment comments", "units": "-"},
        "ManufactureYear": {
            "long_name": "Manufacture year",
            "units": "-",
            "_FillValue": "-1",
        },
        "ManufactureMonth": {
            "long_name": "Manufacture month",
            "units": "-",
            "_FillValue": "-1",
        },
        "ManufactureSensorType": {"long_name": "Manufacture Sensor Type", "units": "-"},
        "ManufactureVoltage": {
            "long_name": "Manufacture voltage",
            "units": "V",
            "_FillValue": "-1",
        },
        "FloatDiameter": {"long_name": "Diameter of surface floater", "units": "cm"},
        "SubsfcFloatPresence": {"long_name": "Subsurface Float Presence", "units": "-"},
        "DrogueType": {"drogue_type": "Drogue Type", "units": "-"},
        "DrogueLength": {"long_name": "Length of drogue.", "units": "m"},
        "DrogueBallast": {
            "long_name": "Weight of the drogue's ballast.",
            "units": "kg",
        },
        "DragAreaAboveDrogue": {"long_name": "Drag area above drogue.", "units": "m^2"},
        "DragAreaOfDrogue": {"long_name": "Drag area drogue.", "units": "m^2"},
        "DragAreaRatio": {"long_name": "Drag area ratio", "units": "m"},
        "DrogueCenterDepth": {
            "long_name": "Average depth of the drogue.",
            "units": "m",
        },
        "DrogueDetectSensor": {"long_name": "Drogue detection sensor", "units": "-"},
        "ve": {"long_name": "Eastward velocity", "units": "m/s"},
        "vn": {"long_name": "Northward velocity", "units": "m/s"},
        "gap": {
            "long_name": "Time interval between previous and next location",
            "units": "s",
        },
        "err_lat": {
            "long_name": "95% confidence interval in latitude",
            "units": "degrees_north",
        },
        "err_lon": {
            "long_name": "95% confidence interval in longitude",
            "units": "degrees_east",
        },
        "err_ve": {
            "long_name": "95% confidence interval in eastward velocity",
            "units": "m/s",
        },
        "err_vn": {
            "long_name": "95% confidence interval in northward velocity",
            "units": "m/s",
        },
        "drogue_status": {
            "long_name": "Status indicating the presence of the drogue",
            "units": "-",
            "flag_values": "1,0",
            "flag_meanings": "drogued, undrogued",
        },
        "sst": {
            "long_name": "Fitted sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated near-surface sea water temperature from drifting buoy measurements. It is the sum of the fitted near-surface non-diurnal sea water temperature and fitted diurnal sea water temperature anomaly. Discrepancies may occur because of rounding.",
        },
        "sst1": {
            "long_name": "Fitted non-diurnal sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated near-surface non-diurnal sea water temperature from drifting buoy measurements",
        },
        "sst2": {
            "long_name": "Fitted diurnal sea water temperature anomaly",
            "units": "Kelvin",
            "comments": "Estimated near-surface diurnal sea water temperature anomaly from drifting buoy measurements",
        },
        "err_sst": {
            "long_name": "Standard uncertainty of fitted sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated one standard error of near-surface sea water temperature estimate from drifting buoy measurements",
        },
        "err_sst1": {
            "long_name": "Standard uncertainty of fitted non-diurnal sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated one standard error of near-surface non-diurnal sea water temperature estimate from drifting buoy measurements",
        },
        "err_sst2": {
            "long_name": "Standard uncertainty of fitted diurnal sea water temperature anomaly",
            "units": "Kelvin",
            "comments": "Estimated one standard error of near-surface diurnal sea water temperature anomaly estimate from drifting buoy measurements",
        },
        "flg_sst": {
            "long_name": "Fitted sea water temperature quality flag",
            "units": "-",
            "flag_values": "0, 1, 2, 3, 4, 5",
            "flag_meanings": "no-estimate, no-uncertainty-estimate, estimate-not-in-range-uncertainty-not-in-range, estimate-not-in-range-uncertainty-in-range estimate-in-range-uncertainty-not-in-range, estimate-in-range-uncertainty-in-range",
        },
        "flg_sst1": {
            "long_name": "Fitted non-diurnal sea water temperature quality flag",
            "units": "-",
            "flag_values": "0, 1, 2, 3, 4, 5",
            "flag_meanings": "no-estimate, no-uncertainty-estimate, estimate-not-in-range-uncertainty-not-in-range, estimate-not-in-range-uncertainty-in-range estimate-in-range-uncertainty-not-in-range, estimate-in-range-uncertainty-in-range",
        },
        "flg_sst2": {
            "long_name": "Fitted diurnal sea water temperature anomaly quality flag",
            "units": "-",
            "flag_values": "0, 1, 2, 3, 4, 5",
            "flag_meanings": "no-estimate, no-uncertainty-estimate, estimate-not-in-range-uncertainty-not-in-range, estimate-not-in-range-uncertainty-in-range estimate-in-range-uncertainty-not-in-range, estimate-in-range-uncertainty-in-range",
        },
    }

    # global attributes
    attrs = {
        "title": "Global Drifter Program hourly drifting buoy collection",
        "history": f"version {GDP_VERSION}. Metadata from dirall.dat and deplog.dat",
        "Conventions": "CF-1.6",
        "time_coverage_start": "",
        "time_coverage_end": "",
        "date_created": datetime.now().isoformat(),
        "publisher_name": "GDP Drifter DAC",
        "publisher_email": "aoml.dftr@noaa.gov",
        "publisher_url": "https://www.aoml.noaa.gov/phod/gdp",
        "license": "freely available",
        "processing_level": "Level 2 QC by GDP drifter DAC",
        "metadata_link": "https://www.aoml.noaa.gov/phod/dac/dirall.html",
        "contributor_name": "NOAA Global Drifter Program",
        "contributor_role": "Data Acquisition Center",
        "institution": "NOAA Atlantic Oceanographic and Meteorological Laboratory",
        "acknowledgement": "Elipot, Shane; Sykulski, Adam; Lumpkin, Rick; Centurioni, Luca; Pazos, Mayra (2022). Hourly location, current velocity, and temperature collected from Global Drifter Program drifters world-wide. [indicate subset used]. NOAA National Centers for Environmental Information. Dataset. https://doi.org/10.25921/x46c-3620. Accessed [date]. Elipot et al. (2022): A Dataset of Hourly Sea Surface Temperature From Drifting Buoys, Scientific Data, 9, 567, https://dx.doi.org/10.1038/s41597-022-01670-2. Elipot et al. (2016): A global surface drifter dataset at hourly resolution, J. Geophys. Res.-Oceans, 121, https://dx.doi.org/10.1002/2016JC011716.",
        "summary": "Global Drifter Program hourly data",
        "doi": "10.25921/x46c-3620",
    }

    # set attributes
    for var in vars_attrs.keys():
        if var in ds.keys():
            ds[var].attrs = vars_attrs[var]
        else:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")
    ds.attrs = attrs

    # rename variables
    ds = ds.rename_vars({"longitude": "lon", "latitude": "lat", "ID": "id"})

    # Cast float64 variables to float32 to reduce memory footprint.
    ds = gdp.cast_float64_variables_to_float32(ds)

    return ds


def to_raggedarray(
    drifter_ids: list[int] | None = None,
    n_random_id: int | None = None,
    url: str = GDP_DATA_URL,
    tmp_path: str | None = None,
) -> RaggedArray:
    """Download and process individual GDP hourly files and return a RaggedArray
    instance with the data.

    Parameters
    ----------
    drifter_ids : list[int], optional
        List of drifters to retrieve (Default: all)
    n_random_id : list[int], optional
        Randomly select n_random_id drifter NetCDF files
    url : str
        URL from which to download the data (Default: GDP_DATA_URL).
        Alternatively, it can be GDP_DATA_URL_EXPERIMENTAL.
    tmp_path : str, optional
        Path to the directory where the individual NetCDF files are stored
        (default varies depending on operating system; /tmp/clouddrift/gdp on Linux)

    Returns
    -------
    out : RaggedArray
        A RaggedArray instance of the requested dataset

    Examples
    --------

    Invoke `to_raggedarray` without any arguments to download all drifter data
    from the 2.01 GDP feed:

    >>> from clouddrift.adapters.gdp1h import to_raggedarray
    >>> ra = to_raggedarray()

    To download a random sample of 100 drifters, for example for development
    or testing, use the `n_random_id` argument:

    >>> ra = to_raggedarray(n_random_id=100)

    To download a specific list of drifters, use the `drifter_ids` argument:

    >>> ra = to_raggedarray(drifter_ids=[44136, 54680, 83463])

    To download the experimental 2.01 GDP feed, use the `url` argument to
    specify the experimental feed URL:

    >>> from clouddrift.adapters.gdp1h import GDP_DATA_URL_EXPERIMENTAL, to_raggedarray
    >>> ra = to_raggedarray(url=GDP_DATA_URL_EXPERIMENTAL)

    Finally, `to_raggedarray` returns a `RaggedArray` instance which provides
    a convenience method to emit a `xarray.Dataset` instance:

    >>> ds = ra.to_xarray()

    To write the ragged array dataset to a NetCDF file on disk, do

    >>> ds.to_netcdf("gdp1h.nc", format="NETCDF4")

    Alternatively, to write the ragged array to a Parquet file, first create
    it as an Awkward Array:

    >>> arr = ra.to_awkward()
    >>> arr.to_parquet("gdp1h.parquet")
    """

    # adjust the tmp_path if using the experimental source
    if tmp_path is None:
        tmp_path = GDP_TMP_PATH if url == GDP_DATA_URL else GDP_TMP_PATH_EXPERIMENTAL

    ids = download(url, tmp_path, drifter_ids, n_random_id)
    filename_pattern = "drifter_hourly_{id}.nc"

    ra = RaggedArray.from_files(
        indices=ids,
        preprocess_func=preprocess,
        name_coords=gdp.GDP_COORDS,
        name_meta=gdp.GDP_METADATA,
        name_data=GDP_DATA,
        name_dims=gdp.GDP_DIMS,
        rowsize_func=gdp.rowsize,
        filename_pattern=filename_pattern,
        tmp_path=tmp_path,
    )

    # set dynamic global attributes
    if ra.attrs_global:
        ra.attrs_global["time_coverage_start"] = (
            f"{datetime(1970, 1, 1) + timedelta(seconds=int(np.min(ra.coords['time']))):%Y-%m-%d:%H:%M:%SZ}"
        )
        ra.attrs_global["time_coverage_end"] = (
            f"{datetime(1970, 1, 1) + timedelta(seconds=int(np.max(ra.coords['time']))):%Y-%m-%d:%H:%M:%SZ}"
        )

    return ra



"""
This module defines functions used to adapt the HURDAT2 cyclone track data as
a ragged-array dataset.
"""

import enum
import os
import re
import tempfile
from dataclasses import dataclass, field, fields
from datetime import datetime, timezone
from io import StringIO
from typing import Literal

import numpy as np
import xarray as xr

from clouddrift.adapters.utils import download_with_progress
from clouddrift.raggedarray import RaggedArray

_DEFAULT_NAME = "hurdat2"
_ATLANTIC_BASIN_URL = "https://www.aoml.noaa.gov/hrd/hurdat/hurdat2.html"
_PACIFIC_BASIN_URL = "https://www.aoml.noaa.gov/hrd/hurdat/hurdat2-nepac.html"

_DEFAULT_FILE_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", _DEFAULT_NAME)

_METERS_IN_NAUTICAL_MILES = 1825
_PASCAL_PER_MILLIBAR = 100

_BasinOption = Literal["atlantic", "pacific", "both"]


class RecordIdentifier(str, enum.Enum):
    """
    C – Closest approach to a coast, not followed by a landfall
    G – Genesis
    I – An intensity peak in terms of both pressure and wind
    L – Landfall (center of system crossing a coastline)
    P – Minimum in central pressure
    R – Provides additional detail on the intensity of the cyclone when rapid changes are underway
    S – Change of status of the system
    T – Provides additional detail on the track (position) of the cyclone
    W – Maximum sustained wind speed
    """

    CLOSEST_TO_COAST = "C"
    GENESIS = "G"
    INTENSITY_PEAK = "I"
    LANDFALL = "L"
    MINIMUM_CENTRAL_PRESSURE = "P"
    RAPID_CHANGE = "R"
    STATUS_CHANGE = "S"
    TRACK_POSITION = "T"
    MAX_SUSTAINED_WIND_SPEED = "W"
    NOT_AVAILABLE = ""


class SystemStatus(str, enum.Enum):
    """
    TD – Tropical cyclone of tropical depression intensity (< 34 knots)
    TS – Tropical cyclone of tropical storm intensity (34-63 knots)
    HU – Tropical cyclone of hurricane intensity (> 64 knots)
    EX – Extratropical cyclone (of any intensity)
    SD – Subtropical cyclone of subtropical depression intensity (< 34 knots)
    SS – Subtropical cyclone of subtropical storm intensity (> 34 knots)
    LO – A low that is neither a tropical cyclone, a subtropical cyclone, nor an extratropical cyclone (of any intensity)
    WV – Tropical Wave (of any intensity)
    DB – Disturbance (of any intensity)
    ET - UNKNOWN found in Northeast Pacific Basin
    PT - UNKNOWN found in Northeast Pacific Basin
    ST - UNKNOWN found in Northeast Pacific Basin
    TY - UNKNOWN found in Northeast Pacific Basin
    """

    TD = "TD"
    TS = "TS"
    HU = "HU"
    EX = "EX"
    SD = "SD"
    SS = "SS"
    LO = "LO"
    WV = "WV"
    DB = "DB"
    ET = "ET"
    PT = "PT"
    ST = "ST"
    TY = "TY"


@dataclass
class HeaderLine:
    id: str = field(
        metadata={
            "standard_name": "automated_tropical_cyclone_forecasting_system_storm_identifier",
            "comment": "The Automated Tropical Cyclone Forecasting System (ATCF) storm identifier is an 8 character string which identifies a tropical cyclone. The storm identifier has the form BBCCYYYY, where BB is the ocean basin, specifically: AL - North Atlantic basin, north of the Equator; SL - South Atlantic basin, south of the Equator; EP - North East Pacific basin, eastward of 140 degrees west longitude; CP - North Central Pacific basin, between the dateline and 140 degrees west longitude; WP - North West Pacific basin, westward of the dateline; IO - North Indian Ocean basin, north of the Equator between 40 and 100 degrees east longitude; SH - South Pacific Ocean basin and South Indian Ocean basin. CC is the cyclone number. Numbers 01 through 49 are reserved for tropical and subtropical cyclones. A cyclone number is assigned to each tropical or subtropical cyclone in each basin as it develops. Numbers are assigned in chronological order. Numbers 50 through 79 are reserved for internal use by operational forecast centers. Numbers 80 through 89 are reserved for training, exercises and testing. Numbers 90 through 99 are reserved for tropical disturbances having the potential to become tropical or subtropical cyclones. The 90's are assigned sequentially and reused throughout the calendar year. YYYY is the four-digit year. This is calendar year for the northern hemisphere. For the southern hemisphere, the year begins July 1, with calendar year plus one. Reference: Miller, R.J., Schrader, A.J., Sampson, C.R., & Tsui, T.L. (1990), The Automated Tropical Cyclone Forecasting System (ATCF), American Meteorological Society Computer Techniques, 5, 653 - 660.",
        }
    )
    basin: str = field(
        metadata={
            "comments": "Basin of origin, possible values: AL - North Atlantic basin, north of the Equator; SL - South Atlantic basin, south of the Equator; EP - North East Pacific basin, eastward of 140 degrees west longitude; CP - North Central Pacific basin, between the dateline and 140 degrees west longitude; WP - North West Pacific basin, westward of the dateline; IO - North Indian Ocean basin, north of the Equator between 40 and 100 degrees east longitude; SH - South Pacific Ocean basin and South Indian Ocean basin",
        }
    )
    year: int = field(metadata={"long_name": "Year"})
    rowsize: int = field(metadata={"comment": "Number of best track entries"})


@dataclass
class DataLine:
    time: datetime = field(
        metadata={"comments": "Computed property from YYY-MM-DD HH:MM in UTC"}
    )
    record_identifier: RecordIdentifier = field(
        metadata={
            "standard_name": "Record Idenfier",
            "comments": RecordIdentifier.__doc__,
        }
    )
    system_status: SystemStatus = field(
        metadata={"standard_name": "System Status", "comments": SystemStatus.__doc__}
    )
    lat: float = field(
        metadata={
            "standard_name": "latitude",
            "units": "degree_north",
        }
    )
    lon: float = field(
        metadata={
            "standard_name": "longitude",
            "units": "degree_east",
        }
    )
    wind_speed: float = field(
        metadata={
            "standard_name": "wind_speed",
            "units": "m s-1",
        }
    )
    pressure: float = field(
        metadata={
            "standard_name": "air_pressure",
            "units": "Pa",
        }
    )
    max_low_wind_radius_ne: float = field(
        metadata={
            "comment": "34 kt Wind maximum radius in the storms NE Quadrant",
            "units": "m",
        }
    )
    max_low_wind_radius_se: float = field(
        metadata={
            "comment": "34 kt Wind maximum radius in the storms SE Quadrant",
            "units": "m",
        }
    )
    max_low_wind_radius_sw: float = field(
        metadata={
            "comment": "34 kt Wind maximum radius in the storms SW Quadrant",
            "units": "m",
        }
    )
    max_low_wind_radius_nw: float = field(
        metadata={
            "comment": "34 kt Wind maximum radius in the storms NW Quadrant",
            "units": "m",
        }
    )
    max_med_wind_radius_ne: float = field(
        metadata={
            "comment": "50 kt Wind maximum radius in the storms NE Quadrant",
            "units": "m",
        }
    )
    max_med_wind_radius_se: float = field(
        metadata={
            "comment": "50 kt Wind maximum radius in the storms SE Quadrant",
            "units": "m",
        }
    )
    max_med_wind_radius_sw: float = field(
        metadata={
            "comment": "50 kt Wind maximum radius in the storms SW Quadrant",
            "units": "m",
        }
    )
    max_med_wind_radius_nw: float = field(
        metadata={
            "comment": "50 kt Wind maximum radius in the storms NW Quadrant",
            "units": "m",
        }
    )
    max_high_wind_radius_ne: float = field(
        metadata={
            "comment": "64 kt Wind maximum radius in the storms NE Quadrant",
            "units": "m",
        }
    )
    max_high_wind_radius_se: float = field(
        metadata={
            "comment": "64 kt Wind maximum radius in the storms SE Quadrant",
            "units": "m",
        }
    )
    max_high_wind_radius_sw: float = field(
        metadata={
            "comment": "64 kt Wind maximum radius in the storms SW Quadrant",
            "units": "m",
        }
    )
    max_high_wind_radius_nw: float = field(
        metadata={
            "comment": "64 kt Wind maximum radius in the storms NW Quadrant",
            "units": "m",
        }
    )
    max_sustained_wind_speed_radius: float = field(
        metadata={
            "standard_name": "radius_of_tropical_cyclone_maximum_sustained_wind_speed",
            "units": "m",
        }
    )


@dataclass
class TrackData:
    global_attrs = {
        "title": "HURricane DATa 2nd generation (HURDAT2)",
        "date_created": datetime.now().isoformat(),
        "publisher_name": "NOAA AOML Hurricane Research Division",
        "publisher_email": "AOML.HRDWebmaster@noaa.gov",
        "publisher_url": "https://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html",
        "institution": "NOAA Atlantic Oceanographic and Meteorological Laboratory",
        "summary": "The National Hurricane Center (NHC) conducts a post-storm analysis of each tropical cyclone in its area of responsibility to determine the official assessment of the cyclone's history",
    }

    header: HeaderLine
    data: list[DataLine]

    def get_rowsize(self) -> int:
        return len(self.data)

    def to_xarray_dataset(self) -> xr.Dataset:
        return xr.Dataset(
            {
                "basin": (["traj"], np.array([self.header.basin])),
                "year": (["traj"], np.array([self.header.year])),
                "rowsize": (["traj"], np.array([self.header.rowsize])),
                "record_identifier": (
                    ["obs"],
                    np.array([line.record_identifier for line in self.data]),
                ),
                "system_status": (
                    ["obs"],
                    np.array([line.system_status for line in self.data]),
                ),
                "lat": (["obs"], np.array([line.lat for line in self.data])),
                "lon": (["obs"], np.array([line.lon for line in self.data])),
                "wind_speed": (
                    ["obs"],
                    np.array([line.wind_speed for line in self.data]),
                ),
                "pressure": (["obs"], np.array([line.pressure for line in self.data])),
                "max_low_wind_radius_ne": (
                    ["obs"],
                    np.array([line.max_low_wind_radius_ne for line in self.data]),
                ),
                "max_low_wind_radius_se": (
                    ["obs"],
                    np.array([line.max_low_wind_radius_se for line in self.data]),
                ),
                "max_low_wind_radius_sw": (
                    ["obs"],
                    np.array([line.max_low_wind_radius_sw for line in self.data]),
                ),
                "max_low_wind_radius_nw": (
                    ["obs"],
                    np.array([line.max_low_wind_radius_nw for line in self.data]),
                ),
                "max_med_wind_radius_ne": (
                    ["obs"],
                    np.array([line.max_med_wind_radius_ne for line in self.data]),
                ),
                "max_med_wind_radius_se": (
                    ["obs"],
                    np.array([line.max_med_wind_radius_se for line in self.data]),
                ),
                "max_med_wind_radius_sw": (
                    ["obs"],
                    np.array([line.max_med_wind_radius_sw for line in self.data]),
                ),
                "max_med_wind_radius_nw": (
                    ["obs"],
                    np.array([line.max_med_wind_radius_nw for line in self.data]),
                ),
                "max_high_wind_radius_ne": (
                    ["obs"],
                    np.array([line.max_high_wind_radius_ne for line in self.data]),
                ),
                "max_high_wind_radius_se": (
                    ["obs"],
                    np.array([line.max_high_wind_radius_se for line in self.data]),
                ),
                "max_high_wind_radius_sw": (
                    ["obs"],
                    np.array([line.max_high_wind_radius_sw for line in self.data]),
                ),
                "max_high_wind_radius_nw": (
                    ["obs"],
                    np.array([line.max_high_wind_radius_nw for line in self.data]),
                ),
                "max_sustained_wind_speed_radius": (
                    ["obs"],
                    np.array(
                        [line.max_sustained_wind_speed_radius for line in self.data]
                    ),
                ),
            },
            coords={
                "id": (["traj"], np.array([self.header.id])),
                "time": (
                    ["obs"],
                    np.array(
                        [int(line.time.timestamp() * 10**9) for line in self.data],
                        dtype="datetime64[ns]",
                    ),
                ),
            },
        )


def _map_heading(coordinate: str) -> float:
    heading_map = {"N": 1, "E": 1, "S": -1, "W": -1}
    cardinal_direction = coordinate[-1]
    heading = coordinate[:-1]

    if (sign := heading_map.get(cardinal_direction)) is not None:
        return float(heading) * sign
    raise ValueError(f"Invalid cardinal direction: {cardinal_direction}")


def _get_download_requests(basin: _BasinOption, tmp_path: str):
    download_requests: list[tuple[str, str, None]] = list()

    if basin == "atlantic" or basin == "both":
        file_name = _ATLANTIC_BASIN_URL.split("/")[-1]
        fp = os.path.join(tmp_path, file_name)
        download_requests.append((_ATLANTIC_BASIN_URL, fp, None))

    if basin == "pacific" or basin == "both":
        file_name = _PACIFIC_BASIN_URL.split("/")[-1]
        fp = os.path.join(tmp_path, file_name)
        download_requests.append((_PACIFIC_BASIN_URL, fp, None))
    return download_requests


def to_raggedarray(
    basin: _BasinOption = "both",
    tmp_path: str = _DEFAULT_FILE_PATH,
    convert: bool = True,
) -> RaggedArray:
    os.makedirs(
        tmp_path, exist_ok=True
    )  # generate temp directory for hurdat related intermerdiary data
    download_requests = _get_download_requests(basin, tmp_path)
    download_with_progress(download_requests)
    track_data = list()

    for _, fp, _ in download_requests:
        track_data.extend(_extract_track_data(fp, convert))

    metadata_fields = list()
    data_fields = list()

    for f in fields(HeaderLine):
        if f.name != "id":
            metadata_fields.append(f.name)

    for f in fields(DataLine):
        if f.name != "time":
            data_fields.append(f.name)

    ra = RaggedArray.from_files(
        indices=list(range(0, len(track_data))),
        name_coords=["id", "time"],
        name_meta=metadata_fields,
        name_data=data_fields,
        name_dims={"traj": "rows", "obs": "obs"},
        rowsize_func=lambda idx: track_data[idx].get_rowsize(),
        preprocess_func=lambda idx: track_data[idx].to_xarray_dataset(),
        attrs_global=TrackData.global_attrs,
        attrs_variables={
            field.name: field.metadata
            for field in fields(HeaderLine) + fields(DataLine)
        },
    )
    return ra


def _apply_or_nan(val: str, cond: bool, func) -> float:
    """If `val` is None or condition is true return nan otherwise apply the `func` to `val` and return the result"""
    if val is None or cond:
        return np.nan
    return func(val)


def _extract_track_data(datafile_path: str, convert: bool) -> list[TrackData]:
    datapage = StringIO()
    with open(datafile_path, "rb") as file:
        while (data := file.read()) != b"":
            datapage.write(data.decode("utf-8").replace(" ", ""))
    datapage.seek(0)

    data_line_count = 0
    current_header = None
    data_lines = list()
    track_data = list[TrackData]()

    is_html_line = lambda line: re.match(r"[html|head|pre]+", line)
    is_header_line = (
        lambda cols, data_line_count: len(cols) == 4 and data_line_count == 0
    )
    is_data_line = lambda cols, data_line_count: len(cols) == 21 and data_line_count > 0
    if convert:
        nm_to_m = (
            lambda x: float(x) * _METERS_IN_NAUTICAL_MILES
        )  # nautical-miles to meters
        k_to_mps = (
            lambda x: float(x) * _METERS_IN_NAUTICAL_MILES / 3600
        )  # knots to meters per second
        mb_to_pa = lambda x: float(x) * _PASCAL_PER_MILLIBAR  # millibar to pascal
    else:
        nm_to_m = lambda x: float(x)
        k_to_mps = lambda x: float(x)
        mb_to_pa = lambda x: float(x)

    while (line := datapage.readline()) != "":
        if is_html_line(line) or line == "\r\n":
            if current_header is not None:
                track_data.append(TrackData(current_header, data_lines))
                data_lines = list()
                current_header = None
            continue

        cols = line.split(",")

        if is_header_line(cols, data_line_count):
            data_line_count = int(cols[2])
            header = HeaderLine(
                id=cols[0],
                basin=cols[0][:2],
                year=int(cols[0][4:8]),
                rowsize=data_line_count,
            )
            if current_header is None:
                current_header = header
            else:
                track_data.append(TrackData(current_header, data_lines))
                data_lines = list()
                current_header = header
        elif (
            is_data_line(cols, data_line_count)
            and len(cols[0]) == 8
            and len(cols[1]) == 4
        ):
            data_lines.append(
                DataLine(
                    time=datetime(
                        year=int(cols[0][:4]),
                        month=int(cols[0][4:6]),
                        day=int(cols[0][6:8]),
                        hour=int(cols[1][:2]),
                        minute=int(cols[1][2:4]),
                        tzinfo=timezone.utc,
                    ),
                    record_identifier=RecordIdentifier(cols[2]),
                    system_status=SystemStatus(cols[3]),
                    lat=_map_heading(cols[4]),
                    lon=_map_heading(cols[5]),
                    wind_speed=_apply_or_nan(cols[6], float(cols[6]) < 0, k_to_mps),
                    pressure=_apply_or_nan(cols[7], cols[7] == "-999", mb_to_pa),
                    max_low_wind_radius_ne=_apply_or_nan(
                        cols[8], cols[8] == "-999", nm_to_m
                    ),
                    max_low_wind_radius_se=_apply_or_nan(
                        cols[9], cols[9] == "-999", nm_to_m
                    ),
                    max_low_wind_radius_sw=_apply_or_nan(
                        cols[10], cols[10] == "-999", nm_to_m
                    ),
                    max_low_wind_radius_nw=_apply_or_nan(
                        cols[11], cols[11] == "-999", nm_to_m
                    ),
                    max_med_wind_radius_ne=_apply_or_nan(
                        cols[12], cols[12] == "-999", nm_to_m
                    ),
                    max_med_wind_radius_se=_apply_or_nan(
                        cols[13], cols[13] == "-999", nm_to_m
                    ),
                    max_med_wind_radius_sw=_apply_or_nan(
                        cols[14], cols[14] == "-999", nm_to_m
                    ),
                    max_med_wind_radius_nw=_apply_or_nan(
                        cols[15], cols[15] == "-999", nm_to_m
                    ),
                    max_high_wind_radius_ne=_apply_or_nan(
                        cols[16], cols[16] == "-999", nm_to_m
                    ),
                    max_high_wind_radius_se=_apply_or_nan(
                        cols[17], cols[17] == "-999", nm_to_m
                    ),
                    max_high_wind_radius_sw=_apply_or_nan(
                        cols[18], cols[18] == "-999", nm_to_m
                    ),
                    max_high_wind_radius_nw=_apply_or_nan(
                        cols[19], cols[19] == "-999", nm_to_m
                    ),
                    max_sustained_wind_speed_radius=_apply_or_nan(
                        cols[20], cols[20] == "-999", nm_to_m
                    ),
                )
            )
            data_line_count -= 1
    return track_data



"""
This module provides functions to calculate various transfer function from wind stress to oceanic
velocity.
"""

import numpy as np
from numpy.lib.scimath import sqrt
from scipy.special import factorial, iv, kv  # type: ignore

from clouddrift.sphere import EARTH_DAY_SECONDS


def slab_wind_transfer(
    omega: float | np.ndarray,
    cor_freq: float,
    friction: float,
    bld: float,
    density: float = 1025.0,
) -> np.ndarray:
    """
    Compute the the transfer function in the case of a ocean slab layer.

    Parameters
    ----------
        omega: array_like or float
            Angular frequency of the wind stress forcing in radians per day.
        cor_freq: float
            Coriolis frequency, in radians per day.
        friction: float
            Friction coefficient, in s-1.
        bld: float
            Thickness of the slab layer, in meters.
        density: float
            Seawater density, in kg m-3. Default is 1025.
    """
    # check that the boundary layer depth is positive
    if bld < 0:
        raise ValueError("Boundary layer depth bld must be positive.")

    # check that the density is positive
    if density < 0:
        raise ValueError("Density density must be positive.")

    # omega can be scalars or arrays, convert to arrays here
    omega_ = np.atleast_1d(omega)

    # convert to radians per second
    omega_ = omega_ / EARTH_DAY_SECONDS
    cor_freq_ = cor_freq / EARTH_DAY_SECONDS

    G = 1 / (density * bld * (friction + 1j * (omega_ + cor_freq_)))

    return G


def wind_transfer(
    omega: float | np.ndarray,
    z: float | np.ndarray,
    cor_freq: float,
    delta: float,
    mu: float,
    bld: float,
    boundary_condition: str = "no-slip",
    method: str = "lilly",
    density: float = 1025.0,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute the transfer function from wind stress to oceanic velocity based on the physically-based
    models of Elipot and Gille (2009) and Lilly and Elipot (2021).

    Parameters
    ----------
        omega: array_like or float
            Angular frequency of the wind stress forcing in radians per day.
        z: array_like or float
            Depth in the ocean, positive downwards, in meters.
        cor_freq: float
            Coriolis frequency, in radians per day.
        delta: float
            Ekman depth, in meters.
        mu: float
            Madsen depth, in meters.
        bld: float
            Boundary layer depth, in meters.
        boundary_condition: str
            Bottom boundary condition at the base of the ocean surface boundary layer.
            Options are "no-slip" (Default) or "free-slip".
        method: str
            Method to compute the transfer function. Options are "lilly" (Default and preferred method)
            or "elipot".
        density: float
            Seawater density, in kg m-3. Default is 1025.

    Returns
    -------
        G: np.ndarray
            The transfer function from wind stress to oceanic velocity in units of kg-1 m 2 s.
        ddelta: np.ndarray
            The gradient of transfer function from the rate of change of the Ekman depth.
        dh: np.ndarray
            The gradient of transfer function from to the rate of change of the boundary layer depth.

    Examples
    --------
        To calculate the transfer function of Lilly and Elipot (2021) corresponding to the Ekman case
         of infinite depth ocean surface boundary layer and constant vertical eddy viscosity, at the zero frequency:

         >>> omega = 0
         >>> z = np.linspace(0, 100, 100)
         >>> cor_freq = 2 * np.pi * 1
         >>> K0 = 1e-4
         >>> delta = np.sqrt(2 * K0 / cor_freq / 86400)
         >>> mu = 0
         >>> G = wind_transfer(omega, z, cor_freq, delta, mu, np.inf, "no-slip")

    Raises
    ------
    ValueError
        If the depth z is not positive.

    References
    ----------
    [1] Elipot, S., and S. T. Gille (2009). Ekman layers in the Southern Ocean: spectral models and
    observations, vertical viscosity and boundary layer depth Ocean Sci., 5, 115–139, 2009, doi:10.5194/os-5-115-2009.

    [2] Lilly, J. M. and S. Elipot (2021). A unifying perspective on transfer function solutions to
    the unsteady Ekman problem. Fluids, 6 (2): 85, 1--36. doi:10.3390/fluids6020085.

    """
    # check that z is positive
    if np.any(z < 0):
        raise ValueError("Depth z must be positive.")

    # check that the boundary layer depth is positive
    if bld < 0:
        raise ValueError("Boundary layer depth bld must be positive.")

    # check that the Ekman depth is positive
    if delta < 0:
        raise ValueError("Ekman depth delta must be positive.")

    # check that the Madsen depth is positive
    if mu < 0:
        raise ValueError("Madsen depth mu must be positive.")

    # check that the density is positive
    if density < 0:
        raise ValueError("Density density must be positive.")

    # check that the boundary condition is valid
    if boundary_condition not in ["no-slip", "free-slip"]:
        raise ValueError("Boundary condition must be 'no-slip' or 'free-slip'.")

    # check that the method is valid
    if method not in ["lilly", "elipot"]:
        raise ValueError("Method must be 'lilly' or 'elipot'.")

    # z and omega can be scalars or arrays, convert to arrays here
    z_ = np.atleast_1d(z)
    omega_ = np.atleast_1d(omega)

    # convert to radians per second
    omega_ = omega_ / EARTH_DAY_SECONDS
    cor_freq_ = cor_freq / EARTH_DAY_SECONDS

    # Create the gridmesh of frequency and depth
    [omega_grid, z_grid] = np.meshgrid(omega_, z_)

    if boundary_condition == "no-slip":
        if method == "lilly":
            G = _wind_transfer_no_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )
        else:
            G = _wind_transfer_elipot_no_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )
    elif boundary_condition == "free-slip":
        if method == "lilly":
            G = _wind_transfer_free_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )
        else:
            G = _wind_transfer_elipot_free_slip(
                omega_grid,
                z_grid,
                cor_freq_,
                delta,
                mu,
                bld,
                density,
            )

    # set G to nan where z > bld; may be mathematcially valid but not physically meaningful
    G[z_grid > bld] = np.nan

    # analytical gradients of the transfer function for mu = 0 and free slip, lilly method
    if boundary_condition == "no-slip" and method == "lilly" and mu == 0:
        s = np.sign(cor_freq_) * np.sign(1 + omega_grid / cor_freq_)
        Gamma = sqrt(2) * _rot(s * np.pi / 4) * sqrt(np.abs(1 + omega_grid / cor_freq_))
        ddelta1 = (
            (Gamma * (bld / delta) * np.tanh(Gamma * (bld / delta)) - 1) * G / delta
        )

        numer = np.exp(Gamma * (-z_grid / delta)) + np.exp(
            -Gamma * (2 * bld - z_grid) / delta
        )
        denom = 1 + np.exp(-Gamma * (2 * bld / delta))
        ddelta2 = (
            -2
            / (delta**2 * np.abs(cor_freq_) * density)
            * (bld - z_grid)
            / delta
            * (numer / denom)
        )

        dG_ddelta = ddelta1 + ddelta2

        dbld1 = -Gamma / delta * np.tanh(Gamma * (bld / delta)) * G
        dbld2 = 2 / (delta**2 * np.abs(cor_freq_) * density) * (numer / denom)
        dG_dbld = dbld1 + dbld2

    else:
        dG_ddelta = np.array([], dtype=float)  # Empty array for consistency
        dG_dbld = np.array([], dtype=float)

    return G, dG_ddelta, dG_dbld


def _wind_transfer_no_slip(
    omega_grid: np.ndarray,
    z_grid: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary condition.
    """

    zo = np.divide(delta**2, mu)
    s = np.sign(coriolis_frequency) * np.sign(1 + omega_grid / coriolis_frequency)

    xiz, xih, xi0 = _xis(s, zo, delta, z_grid, omega_grid, coriolis_frequency, bld)

    if bld == np.inf:
        if mu == 0:
            # Ekman solution
            coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                delta * np.abs(coriolis_frequency) * density
            )
            G = (
                coeff
                * (
                    np.exp(
                        -(1 + s * 1j)
                        * (z_grid / delta)
                        * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
                    )
                )
                / (sqrt(np.abs(1 + omega_grid / coriolis_frequency)))
            )
        elif delta == 0:
            # Madsen solution
            coeff = 4 / (density * np.abs(coriolis_frequency) * mu)
            G = coeff * kv(
                0,
                2
                * sqrt(2)
                * _rot(s * np.pi / 4)
                * sqrt((z_grid / mu) * np.abs(1 + omega_grid / coriolis_frequency)),
            )
        else:
            # mixed solution
            k0z = kv(0, xiz)
            k10 = kv(1, xi0)
            coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                delta
                * np.abs(coriolis_frequency)
                * density
                * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
            )
            G = coeff * k0z / k10
    else:
        if mu == 0:
            # finite layer Ekman
            coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
                delta
                * np.abs(coriolis_frequency)
                * density
                * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
            )
            argh = (
                sqrt(2)
                * _rot(s * np.pi / 4)
                * (bld / delta)
                * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
            )
            argz = (
                sqrt(2)
                * _rot(s * np.pi / 4)
                * (z_grid / delta)
                * sqrt(np.abs(1 + omega_grid / coriolis_frequency))
            )

            numer = np.exp(-argz) - np.exp(argz) * np.exp(-2 * argh)
            denom = 1 + np.exp(-2 * argh)
            G = coeff * np.divide(numer, denom)

            # solution at inertial frequency
            bool_idx = omega_grid == -coriolis_frequency
            G[bool_idx] = 2 / (
                (density * np.abs(coriolis_frequency) * delta**2)
                * (bld - z_grid[bool_idx])
            )

        elif delta == 0:
            # finite layer Madsen
            coeff = 4 / (density * np.abs(coriolis_frequency) * mu)
            argz = (
                2
                * sqrt(2)
                * _rot(s * np.pi / 4)
                * sqrt((z_grid / mu) * np.abs(1 + omega_grid / coriolis_frequency))
            )
            argh = (
                2
                * sqrt(2)
                * _rot(s * np.pi / 4)
                * sqrt((bld / mu) * np.abs(1 + omega_grid / coriolis_frequency))
            )
            k0z, i0z, k0h, i0h, _, _ = _bessels_noslip(argz, argh)
            G = coeff * (k0z - i0z * k0h / i0h)

            # solution at inertial frequency
            bool_idx = omega_grid == -coriolis_frequency
            if isinstance(z_grid, np.ndarray):
                G[bool_idx] = 0.5 * coeff * np.log(bld / z_grid[bool_idx])
            else:
                G = 0.5 * coeff * np.log(bld / z_grid)
        else:
            # finite layer mixed
            G = _wind_transfer_general_no_slip(
                omega_grid, z_grid, coriolis_frequency, delta, mu, bld, density
            )

    return G


def _wind_transfer_general_no_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary condition, general form.
    """
    zo = np.divide(delta**2, mu)
    s = np.sign(coriolis_frequency) * np.sign(1 + omega / coriolis_frequency)
    xiz, xih, xi0 = _xis(s, zo, delta, z, omega, coriolis_frequency, bld)
    coeff = (
        sqrt(2)
        * _rot(-s * np.pi / 4)
        / (
            delta
            * density
            * np.abs(coriolis_frequency)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
    )

    k0z, i0z, k0h, i0h, k10, i10 = _bessels_noslip(xiz, xih, xi0=xi0)
    numer = k0z / k10 - (k0h / k10) * (i0z / i0h)
    denom = 1 + (k0h / k10) * (i10 / i0h)
    G = coeff * np.divide(numer, denom)

    # large argument approximation
    bool_idx = np.log10(np.abs(xiz)) > 2.9
    G[bool_idx] = _wind_transfer_general_no_slip_expansion(
        omega[bool_idx], z[bool_idx], coriolis_frequency, delta, mu, bld, density
    )

    # inertial limit
    G = _wind_transfer_inertiallimit(
        G, omega, z, coriolis_frequency, delta, mu, bld, density
    )

    return G


def _wind_transfer_general_no_slip_expansion(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Compute the transfer function from wind stress to oceanic velocity with no-slip boundary, large argument approximation.
    """
    zo = np.divide(delta**2, mu)
    s = np.sign(coriolis_frequency) * np.sign(1 + omega / coriolis_frequency)
    xiz, xih, xi0 = _xis(s, zo, delta, z, omega, coriolis_frequency, bld)
    coeff = (
        sqrt(2)
        * _rot(-s * np.pi / 4)
        / (
            delta
            * density
            * np.abs(coriolis_frequency)
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
    )
    k0z, i0z, k0h, i0h, k10, i10 = _besseltildes_noslip(xiz, xih, xi0, 30)

    numer = np.exp(xi0 - xiz) * i0h * k0z - np.exp(xi0 + xiz - 2 * xih) * k0h * i0z
    denom = i0h * k10 + np.exp(2 * xi0 - 2 * xih) * k0h * i10
    G = coeff * np.divide(numer, denom)

    bool_idx = omega == -coriolis_frequency
    G[bool_idx] = (
        4
        * zo
        / (
            density
            * np.abs(coriolis_frequency)
            * delta**2
            * np.divide(sqrt(1 + bld / zo) - sqrt(1 + z / zo), (1 + z / zo) ** 0.25)
        )
    )

    return G


def _wind_transfer_inertiallimit(
    G: np.ndarray,
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_freq: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary, inertial limit.
    """
    zo = delta**2 / mu
    bool_idx = omega == -coriolis_freq

    if np.any(bool_idx):
        if not np.isinf(bld) and not np.isinf(zo):
            G[bool_idx] = (2 / (density * np.abs(coriolis_freq) * mu)) * np.log(
                (1 + bld / zo) / (1 + z[bool_idx] / zo)
            )
        elif not np.isinf(bld) and np.isinf(zo):
            G[bool_idx] = (2 / (density * np.abs(coriolis_freq) * delta**2)) * (
                bld - z[bool_idx]
            )
        else:
            G[bool_idx] = np.inf

    return G


def _wind_transfer_free_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with free-slip boundary condition.
    """

    zo = np.divide(delta**2, mu)

    s = np.sign(coriolis_frequency) * np.sign(1 + omega / coriolis_frequency)

    xiz, xih, xi0 = _xis(s, zo, delta, z, omega, coriolis_frequency, bld)

    if delta != 0.0 and mu != 0.0:
        coeff = (
            sqrt(2)
            * _rot(-s * np.pi / 4)
            / (
                delta
                * density
                * np.abs(coriolis_frequency)
                * sqrt(np.abs(1 + omega / coriolis_frequency))
            )
        )
        k0z, i0z, k1h, i1h, k10, i10 = _bessels_freeslip(xiz, xih, xi0=xi0)
        numer = i0z * k1h + i1h * k0z
        denom = i1h * k10 - i10 * k1h
        G = coeff * np.divide(numer, denom)

    elif mu == 0.0:
        coeff = (sqrt(2) * _rot(-s * np.pi / 4)) / (
            delta
            * np.abs(coriolis_frequency)
            * density
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        cosharg = (
            sqrt(2)
            * _rot(s * np.pi / 4)
            * (bld - z)
            / delta
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        sinharg = (
            sqrt(2)
            * _rot(s * np.pi / 4)
            * bld
            / delta
            * sqrt(np.abs(1 + omega / coriolis_frequency))
        )
        G = coeff * np.cosh(cosharg) / np.sinh(sinharg)

    elif delta == 0.0:
        k0z, i0z, k1h, i1h, _, _ = _bessels_freeslip(xiz, xih)

        K1 = 0.5 * mu * np.abs(coriolis_frequency)

        coeff = 2 / (density * K1)

        G = coeff * (k0z + k1h * i0z / i1h)

    return G


def _wind_transfer_elipot_no_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with no-slip boundary condition, Elipot method.
    """

    zo = np.divide(delta**2, mu)
    K0 = 0.5 * delta**2 * np.abs(coriolis_frequency)
    K1 = 0.5 * mu * np.abs(coriolis_frequency)

    delta1 = sqrt(2 * K0 / (omega + coriolis_frequency))
    delta2 = K1 / (omega + coriolis_frequency)
    xiz = 2 * sqrt(1j * (zo + z) / delta2)
    xi0 = 2 * sqrt(1j * zo / delta2)
    xih = 2 * sqrt(1j * (zo + bld) / delta2)

    if bld == np.inf:
        if K1 == 0:
            # Ekman solution
            coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
            G = coeff * np.exp(-z * (1 + 1j) / delta1)
        elif K0 == 0:
            # Madsen solution
            coeff = 2 / (density * K1)
            k0z = kv(0, 2 * sqrt(1j * z / delta2))
            G = coeff * k0z
        else:
            # Mixed solution
            coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
            G = coeff * kv(0, xiz) / kv(1, xi0)
    else:
        if K1 == 0:
            # Finite-layer Ekman
            coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
            numer = np.sinh((1 + 1j) * (bld - z) / delta1)
            denom = np.cosh((1 + 1j) * bld / delta1)
            G = coeff * numer / denom
        elif K0 == 0:
            # Finite-layer Madsen
            coeff = 2 / (density * K1)
            k0z = kv(0, 2 * sqrt(1j * z / delta2))
            k0h = kv(0, 2 * sqrt(1j * bld / delta2))
            i0z = iv(0, 2 * sqrt(1j * z / delta2))
            i0h = iv(0, 2 * sqrt(1j * bld / delta2))
            G = coeff * (k0z - (k0h * i0z / i0h))
        else:
            # Finite-layer mixed solution
            coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
            k0z, i0z, k0h, i0h, k10, i10 = _bessels_noslip(xiz, xih, xi0)
            numer = i0h * k0z - k0h * i0z
            denom = i10 * k0h + k10 * i0h
            G = coeff * numer / denom

    return G


def _wind_transfer_elipot_free_slip(
    omega: np.ndarray,
    z: np.ndarray,
    coriolis_frequency: float,
    delta: float,
    mu: float,
    bld: float,
    density: float,
) -> np.ndarray:
    """
    Transfer function from wind stress to oceanic velocity with free-slip boundary condition, Elipot method.
    """

    zo = np.divide(delta**2, mu)
    K0 = 0.5 * delta**2 * np.abs(coriolis_frequency)
    K1 = 0.5 * mu * np.abs(coriolis_frequency)
    delta1 = sqrt(2 * K0 / (omega + coriolis_frequency))
    delta2 = K1 / (omega + coriolis_frequency)
    xiz = 2 * sqrt(1j * (zo + z) / delta2)
    xi0 = 2 * sqrt(1j * zo / delta2)
    xih = 2 * sqrt(1j * (zo + bld) / delta2)

    if K1 == 0.0:
        # Finite-layer Ekman
        coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
        numer = np.cosh((1 + 1j) * (bld - z) / delta1)
        denom = np.sinh((1 + 1j) * bld / delta1)
        G = coeff * numer / denom
    elif K0 == 0.0:
        # Finite-layer Madsen
        coeff = 2 / (density * K1)
        k0z = kv(0, 2 * sqrt(1j * z / delta2))
        k0h = kv(0, 2 * sqrt(1j * bld / delta2))
        i0z = iv(0, 2 * sqrt(1j * z / delta2))
        i1h = iv(1, 2 * sqrt(1j * bld / delta2))
        G = coeff * (k0z - (k0h * i0z / i1h))
    else:
        # Finite-layer mixed solution
        coeff = 1 / (density * sqrt(1j * (omega + coriolis_frequency) * K0))
        k1h = kv(1, xih)
        i0z = iv(0, xiz)
        i1h = iv(1, xih)
        k10 = kv(1, xi0)
        i10 = iv(1, xi0)
        k0z = kv(0, xiz)
        numer = k1h * i0z + i1h * k0z
        denom = i1h * k10 - i10 * k1h
        G = coeff * numer / denom

    return G


def _bessels_freeslip(
    xiz: float | np.ndarray,
    xih: float | np.ndarray,
    xi0: float | np.ndarray | None | None = None,
) -> tuple[
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
]:
    """
    Compute the Bessel functions for the free-slip boundary condition for the xsi(z), xsi(h), and xsi(0) functions.
    """
    k0z = kv(0, xiz)
    i0z = iv(0, xiz)
    k1h = kv(1, xih)
    i1h = iv(0, xih)

    if xi0 is not None:
        k10 = kv(1, xi0)
        i10 = iv(1, xi0)
        return k0z, i0z, k1h, i1h, k10, i10
    else:
        return k0z, i0z, k1h, i1h, np.nan, np.nan


def _bessels_noslip(
    xiz: float | np.ndarray,
    xih: float | np.ndarray,
    xi0: float | np.ndarray | None = None,
) -> tuple[
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
]:
    """
    Compute the Bessel functions for the no-slip boundary condition for the xsi(z), xsi(h), and xsi(0) functions.
    """
    k0z = kv(0, xiz)
    i0z = iv(0, xiz)
    k0h = kv(0, xih)
    i0h = iv(0, xih)

    if xi0 is not None:
        k10 = kv(1, xi0)
        i10 = iv(1, xi0)
        return k0z, i0z, k0h, i0h, k10, i10
    else:
        return k0z, i0z, k0h, i0h, np.nan, np.nan


def _besseltildes_noslip(
    xiz: float | np.ndarray,
    xih: float | np.ndarray,
    xi0: float | np.ndarray,
    nterms: int = 30,
) -> tuple[
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
    float | np.ndarray,
]:
    """
    Compute the n-term expansion about the large-argument exponential behavior of Bessel functions
    for the xsi(z), xsi(h), and xsi(0) functions.
    """
    k0z = kvtilde(0, xiz, nterms)
    i0z = ivtilde(0, xiz, nterms)
    k0h = kvtilde(0, xih, nterms)
    i0h = ivtilde(0, xih, nterms)
    k10 = kvtilde(1, xi0, nterms)
    i10 = ivtilde(1, xi0, nterms)

    return k0z, i0z, k0h, i0h, k10, i10


def kvtilde(
    nu: int,
    z: float | np.ndarray,
    nterms: int = 30,
) -> float | np.ndarray:
    """
    Compute the n-term expansion about the large-argument exponential behavior of the modified Bessel
    function of the second kind of real order (kv).
    """
    z = np.asarray(
        z, dtype=np.complex128
    )  # Ensure z is an array for vectorized operations
    sizez = z.shape
    z = z.ravel()

    # Prepare zk for vectorized computation
    zk = np.tile(z, (nterms, 1)).T
    zk[:, 0] = 1.0
    zk = np.cumprod(zk, axis=1)

    k = np.arange(nterms)
    ak = 4.0 * nu**2 - (2.0 * k - 1.0) ** 2
    ak[0] = 1.0
    ak = np.cumprod(ak) / (factorial(k) * (8.0**k))

    # Handling potential non-finite values in ak
    if not np.all(np.isfinite(ak)):
        first_nonfinite = np.where(~np.isfinite(ak))[0][0]
        ak = ak[:first_nonfinite]
        zk = zk[:, :first_nonfinite]

    K = sqrt(np.pi / (2.0 * z)) * (np.dot(1.0 / zk, ak))
    K = K.reshape(sizez)

    return K


def ivtilde(
    nu: int,
    z: float | np.ndarray,
    nterms: int = 30,
) -> float | np.ndarray:
    """
    Compute the n-term expansion about the large-argument exponential behavior of the
    Modified Bessel function of the first kind of real order (iv).
    """
    z = np.asarray(
        z, dtype=np.complex128
    )  # Ensure z is an array for vectorized operations
    sizez = z.shape
    z = z.ravel()

    # Prepare zk for vectorized computation with alternating signs for each term
    zk = np.tile(-z, (nterms, 1)).T
    zk[:, 0] = 1.0
    zk = np.cumprod(zk, axis=1)

    k = np.arange(nterms)
    ak = 4.0 * nu**2 - (2.0 * k - 1.0) ** 2
    ak[0] = 1.0
    ak = np.cumprod(ak) / (factorial(k) * (8.0**k))

    # Handling potential non-finite values in ak
    if not np.all(np.isfinite(ak)):
        first_nonfinite = np.where(~np.isfinite(ak))[0][0]
        ak = ak[:first_nonfinite]
        zk = zk[:, :first_nonfinite]

    I = 1.0 / sqrt(2.0 * z * np.pi) * (np.dot(1.0 / zk, ak))
    I = I.reshape(sizez)

    return I


def _xis(
    s: float,
    zo: float,
    delta: float,
    z: float | np.ndarray,
    omega: float | np.ndarray,
    coriolis_frequency: float,
    bld: float,
) -> tuple[float | np.ndarray, float | np.ndarray, float | np.ndarray]:
    """
    Compute the complex-valued "xsi" functions.
    """
    xiz = (
        2
        * sqrt(2)
        * _rot(s * np.pi / 4)
        * np.divide(zo, delta)
        * sqrt((1 + np.divide(z, zo)) * np.abs(1 + omega / coriolis_frequency))
    )
    xih = (
        2
        * sqrt(2)
        * _rot(s * np.pi / 4)
        * np.divide(zo, delta)
        * sqrt((1 + np.divide(bld, zo)) * np.abs(1 + omega / coriolis_frequency))
    )
    xi0 = (
        2
        * sqrt(2)
        * _rot(s * np.pi / 4)
        * np.divide(zo, delta)
        * sqrt(np.abs(1 + omega / coriolis_frequency))
    )

    return xiz, xih, xi0


def _rot(x: float) -> complex:
    """
    Compute the complex-valued rotation.
    """
    return np.exp(1j * x)



import os
import shutil

import numpy as np

import tests.utils as testutils
from clouddrift.adapters import gdp6h


class gdp6h_integration_tests(testutils.DisableProgressTestCase):
    def test_load_subset_and_create_aggregate(self):
        ra = gdp6h.to_raggedarray(n_random_id=5, tmp_path=gdp6h.GDP_TMP_PATH)
        assert "rowsize" in ra.metadata
        assert "temp" in ra.data
        assert "ve" in ra.data
        assert "vn" in ra.data
        assert ra.coords["id"].dtype == np.int64
        assert len(ra.data["vn"]) == len(ra.coords["time"])
        assert len(ra.metadata["rowsize"]) == len(ra.coords["id"])

        agg_path = os.path.join(gdp6h.GDP_TMP_PATH, "aggregate")
        os.makedirs(agg_path, exist_ok=True)
        ra.to_netcdf(os.path.join(agg_path, "gdp6h_5r_sample.nc"))

    @classmethod
    def tearDownClass(cls):
        [shutil.rmtree(dir) for dir in [gdp6h.GDP_TMP_PATH]]



import unittest
from datetime import datetime, timedelta
from io import BufferedIOBase
from unittest.mock import Mock, call, mock_open, patch

from requests import RequestException

from clouddrift.adapters import utils
from tests.adapters.utils import MultiPatcher


class utils_tests(unittest.TestCase):
    head_response_mock: Mock
    get_response_mock: Mock
    requests_mock: Mock
    open_mock: Mock
    bar_mock: Mock

    def setUp(self) -> None:
        """
        Set up the mocks for the tests.
        """

        # Mock responses for requests.head and requests.get
        self.head_response_mock = Mock()
        self.head_response_mock.headers = {"Last-Modified": "some-date-str"}
        self.head_response_mock.__enter__ = Mock(return_value=self.head_response_mock)
        self.head_response_mock.__exit__ = Mock()

        self.get_response_mock = Mock()
        self.get_response_mock.headers = {"Content-Length": "4"}
        self.get_response_mock.iter_content = Mock(
            return_value=[b"a", b"b", b"c", b"d"]
        )
        self.get_response_mock.__enter__ = Mock(return_value=self.get_response_mock)
        self.get_response_mock.__exit__ = Mock()

        # Create a mock requests module
        self.requests_mock = Mock()
        self.requests_mock.head = Mock(return_value=self.head_response_mock)
        self.requests_mock.get = Mock(return_value=self.get_response_mock)
        self.requests_mock.RequestException = RequestException  # Assign real exception

        # Mock open
        self.open_mock = mock_open()

        # Mock tqdm progress bar
        self.bar_mock = Mock()
        self.bar_mock.update = Mock()

        # Patch 'requests' in 'clouddrift.adapters.utils' with 'self.requests_mock'
        self.requests_patcher = patch(
            "clouddrift.adapters.utils.requests", self.requests_mock
        )
        self.requests_patcher.start()

        # Patch 'open' in 'clouddrift.adapters.utils' with 'self.open_mock'
        self.open_patcher = patch("clouddrift.adapters.utils.open", self.open_mock)
        self.open_patcher.start()

        # Ensure patches are stopped after tests
        self.addCleanup(self.requests_patcher.stop)
        self.addCleanup(self.open_patcher.stop)

    def test_forgo_download_no_update(self):
        """
        Don't download from the server if we have the latest version of the file.
        """

        now = datetime.now()
        datetime_mock = Mock()
        datetime_mock.strptime = Mock(return_value=now)

        with MultiPatcher(
            [
                patch("clouddrift.adapters.utils.datetime", datetime_mock),
                patch(
                    "clouddrift.adapters.utils.os.path.getmtime",
                    Mock(return_value=now.timestamp()),
                ),
                patch("clouddrift.adapters.utils.requests", self.requests_mock),
                patch("clouddrift.adapters.utils.open", self.open_mock),
                patch(
                    "clouddrift.adapters.utils.os.path.exists", Mock(return_value=True)
                ),
            ]
        ) as _:
            utils._download_with_progress(
                "some.url.com", "./some/path/existing-file.nc", 0, False
            )
            self.requests_mock.get.assert_not_called()

    def test_download_new_update_buffer(self):
        """
        Download the file from the server if there is an update to it remotely (BufferedIOBase).
        """

        now = datetime.now()
        datetime_mock = Mock()
        datetime_mock.strptime = Mock(return_value=now)
        with MultiPatcher(
            [
                patch("clouddrift.adapters.utils.datetime", datetime_mock),
                patch(
                    "clouddrift.adapters.utils.os.path.exists", Mock(return_value=True)
                ),
                patch(
                    "clouddrift.adapters.utils.os.path.getmtime",
                    Mock(return_value=(now - timedelta(days=1)).timestamp()),
                ),
                patch("clouddrift.adapters.utils.os.remove", Mock()),
                patch("clouddrift.adapters.utils.os.rename", Mock()),
            ]
        ):
            buffer = Mock(spec=BufferedIOBase)
            utils._download_with_progress("some.url.com", buffer, 0, False)

            # Assertions to verify the behavior
            self.requests_mock.get.assert_called_with(
                "some.url.com", timeout=10, stream=True
            )
            buffer.write.assert_has_calls(
                [call(b"a"), call(b"b"), call(b"c"), call(b"d")]
            )

    def test_download_new_update_file(self):
        """
        Download the file from the server if there is an update to it remotely (file path).
        """

        now = datetime.now()
        datetime_mock = Mock()
        datetime_mock.strptime = Mock(return_value=now)
        with MultiPatcher(
            [
                patch("clouddrift.adapters.utils.datetime", datetime_mock),
                patch(
                    "clouddrift.adapters.utils.os.path.exists", Mock(return_value=True)
                ),
                patch(
                    "clouddrift.adapters.utils.os.path.getmtime",
                    Mock(return_value=(now - timedelta(days=1)).timestamp()),
                ),
                patch(
                    "clouddrift.adapters.utils.os.path.getsize", Mock(return_value=4)
                ),
                patch("clouddrift.adapters.utils.os.remove", Mock()),
                patch("clouddrift.adapters.utils.os.rename", Mock()),
            ]
        ):
            output_file = "output.file"
            utils._download_with_progress("some.url.com", output_file, 0, False)

            # Assertions to verify the behavior
            self.requests_mock.get.assert_called_with(
                "some.url.com", timeout=10, stream=True
            )
            self.open_mock.assert_called_with(output_file + ".part", "wb")
            handle = self.open_mock()
            handle.write.assert_has_calls(
                [call(b"a"), call(b"b"), call(b"c"), call(b"d")]
            )

    def test_progress_mechanism_disabled_files(self):
        """
        Ensure we don't show progress for the list of files when number of files is less than 20 and user
        does not specify whether the progress feature is enabled/disabled.
        """

        mocked_futures = [self.gen_future_mock() for _ in range(0, 3)]
        download_requests = [
            ("src0", "dst"),
            ("src1", "dst"),
            ("src2", "dst"),
        ]

        tpe_mock = Mock()
        tpe_mock.__enter__ = Mock(return_value=tpe_mock)
        tpe_mock.__exit__ = Mock()
        tpe_mock.submit = Mock(side_effect=mocked_futures)

        futures_mock = Mock()
        futures_mock.ThreadPoolExecutor = Mock(return_value=tpe_mock)
        futures_mock.as_completed = Mock(return_value=mocked_futures)

        with MultiPatcher(
            [
                patch(
                    "clouddrift.adapters.utils.tqdm", Mock(return_value=self.bar_mock)
                ),
                patch("clouddrift.adapters.utils.concurrent.futures", futures_mock),
                patch("clouddrift.adapters.utils.os.remove", Mock()),
                patch("clouddrift.adapters.utils.os.rename", Mock()),
            ]
        ) as mock:
            utils.download_with_progress(download_requests)
            mock[0].assert_not_called()
            self.bar_mock.update.assert_not_called()

    def test_progress_mechanism_enabled_files(self):
        """
        Ensure we do show progress for the list of files when number of files is greater than 20 and user
        does not specify whether the progress feature is enabled/disabled.
        """

        mocked_futures = [self.gen_future_mock() for _ in range(0, 21)]
        download_requests = [("src0", "dst", None) for _ in range(0, 21)]

        tpe_mock = Mock()
        tpe_mock.__enter__ = Mock(return_value=tpe_mock)
        tpe_mock.__exit__ = Mock()
        tpe_mock.submit = Mock(side_effect=mocked_futures)

        futures_mock = Mock()
        futures_mock.ThreadPoolExecutor = Mock(return_value=tpe_mock)
        futures_mock.as_completed = Mock(return_value=mocked_futures)

        with MultiPatcher(
            [
                patch(
                    "clouddrift.adapters.utils.tqdm", Mock(return_value=self.bar_mock)
                ),
                patch("clouddrift.adapters.utils.concurrent.futures", futures_mock),
                patch("clouddrift.adapters.utils.os.remove", Mock()),
                patch("clouddrift.adapters.utils.os.rename", Mock()),
            ]
        ) as mock:
            utils.download_with_progress(download_requests)
            mock[0].assert_called()
            self.bar_mock.update.assert_called()

    def test_progress_mechanism_enabled_file(self):
        """
        Ensure we do show progress per file when the user enables the feature.
        """

        tqdm_mock = Mock(return_value=self.bar_mock)
        with MultiPatcher(
            [
                patch("clouddrift.adapters.utils.tqdm", tqdm_mock),
                patch(
                    "clouddrift.adapters.utils.os.path.exists", Mock(return_value=False)
                ),
                patch(
                    "clouddrift.adapters.utils.os.path.getsize", Mock(return_value=4)
                ),
                patch("clouddrift.adapters.utils.os.remove", Mock()),
                patch("clouddrift.adapters.utils.os.rename", Mock()),
            ]
        ):
            output_file = "output.file"
            utils._download_with_progress("some.url.com", output_file, 0, True)

            # Assertions
            tqdm_mock.assert_called_once()
            self.bar_mock.update.assert_has_calls([call(1), call(1), call(1), call(1)])
            self.requests_mock.get.assert_called_with(
                "some.url.com", timeout=10, stream=True
            )
            self.open_mock.assert_called_with(output_file + ".part", "wb")

    def test_progress_mechanism_disabled_file(self):
        """
        Ensure we don't show progress per file when the user disables the feature.
        """

        tqdm_mock = Mock(return_value=self.bar_mock)
        with MultiPatcher(
            [
                patch("clouddrift.adapters.utils.tqdm", tqdm_mock),
                patch(
                    "clouddrift.adapters.utils.os.path.exists", Mock(return_value=False)
                ),
                patch(
                    "clouddrift.adapters.utils.os.path.getsize", Mock(return_value=4)
                ),
                patch("clouddrift.adapters.utils.os.remove", Mock()),
                patch("clouddrift.adapters.utils.os.rename", Mock()),
            ]
        ):
            output_file = "output.file"
            utils._download_with_progress("some.url.com", output_file, 0, False)

            # Assertions
            tqdm_mock.assert_not_called()
            self.bar_mock.update.assert_not_called()
            self.requests_mock.get.assert_called_with(
                "some.url.com", timeout=10, stream=True
            )
            self.open_mock.assert_called_with(output_file + ".part", "wb")

    def test_raises_on_any_exception_and_cleanup(self):
        """
        Ensure that any download job exception is propogated up and all download jobs are cancelled and cleaned up.
        """

        tpe_mock = Mock()
        tpe_mock.shutdown = Mock()
        tpe_mock.__enter__ = Mock(return_value=tpe_mock)
        tpe_mock.__exit__ = Mock()

        mocked_futures = [
            self.gen_future_mock(),
            self.gen_future_mock(),
            self.gen_future_mock(Exception("just a test exception that is expected")),
            self.gen_future_mock(complete=True),
        ]

        futures_mock = Mock()
        futures_mock.ThreadPoolExecutor = Mock(return_value=tpe_mock)
        futures_mock.as_completed = Mock(return_value=mocked_futures)
        tpe_mock.submit = Mock(side_effect=mocked_futures)

        os_mock = Mock()
        os_mock.remove = Mock()

        with MultiPatcher(
            [
                patch(
                    "clouddrift.adapters.utils.tqdm", Mock(return_value=self.bar_mock)
                ),
                patch("clouddrift.adapters.utils.concurrent.futures", futures_mock),
                patch("clouddrift.adapters.utils.os", os_mock),
            ]
        ) as _:
            self.assertRaises(
                Exception,
                utils.download_with_progress,
                [("src", "dst", None) for _ in range(0, len(mocked_futures))],
                show_list_progress=True,
            )
            assert tpe_mock.submit.call_count == len(mocked_futures)
            assert self.bar_mock.update.call_count == 2
            assert os_mock.remove.call_count == len(mocked_futures)
            tpe_mock.shutdown.assert_called_once()
            [fut_mock.cancel.assert_called_once() for fut_mock in mocked_futures[:-1]]
            mocked_futures[-1].cancel.assert_not_called()

    def gen_future_mock(self, ex=None, complete=False):
        future = Mock()
        future.exception = Mock(return_value=ex)
        future.done = Mock(return_value=complete)
        future.cancel = Mock()
        return future
